#
# See the file "LICENSE" for the full license governing this code.
#
# frozen_string_literal: true
require 'easy_type/catalog_data'
require 'easy_type/classifier'
require 'easy_type/pdb'

module EasyType
  #
  # Class to read and write from the NodeGroup cache
  #
  class NodeGroupData
    include EasyType::CatalogData
    include EasyType::Encryption

    def initialize(cache_file)
      @cache_file = cache_file
    end

    #
    # Load the node group information from the catalog. If there is np catalog available, 
    # load the node group information from the cache
    #
    def load_from_catalog
      @data = get_secured_catalog_data('node_group_data')
    end

    #
    # Load the node group information from the puppet server
    #
    def load_from_server
      Puppet.debug "Fetching node group information from classifier."
      classifier = EasyType::Classifier.new
      groups = classifier.em_groups
      @data = {}
      if groups.nil?
        Puppet.debug "Node group '#{group_name}' not found."
      else
        groups.each {|g| get_nodes(g, @data)}
      end
      @data
    end

    #
    # This function fetches the node group data from the cache file. If the
    # node group data is older then 30 minutes, we fail and fall back to 
    # an empty list. Thus failing the license checks.
    #
    def load_from_cache
      encrypted_data = File.read(@cache_file) 
      node_group_data = JSON.parse(decrypt(encrypted_data))
      if Time.parse(node_group_data['time']) < Time.now - 90 * 60  # Older then 1,5 hour
        Puppet.warn_once(:license, :node_group_cache_expired, "Node group cache older the 1,5 hour. Please run puppet agent to update the license information.")
        @data = {}
      else
        Puppet.debug "Using cached node group data."
        @data = node_group_data['data']
      end
    rescue
      Puppet.debug "Error loading from cache, returning empty list."
      @data = {}
    end

    #
    # This function saves the fetched node group data to the node group cache file.
    # It also stores the time the cache is created so when we use it, we can check how old it is.
    # 
    # Because it is called multiple times. We only rewrite the file if it is older then 1 minute.
    #
    def save_node_group_cache
      dir = File.dirname(@cache_file)
      FileUtils.mkdir_p(dir) unless File.exist?(dir)
      if !File.exist?(@cache_file) || (File.exist?(@cache_file) && File.ctime(@cache_file) < Time.now - 60)
        hashed_data = {
          :time => Time.now,
          :data => @data
        }
        encrypted_data = encrypt(hashed_data.to_json)
        File.write(@cache_file,encrypted_data) 
      end
    end

    private
  
    def get_nodes(group, data)
      full_name = group.full_name.gsub('/All Nodes/EM licensed nodes/','')
      return if full_name == '/All Nodes/EM licensed nodes'
      module_name, license_id = full_name.split('/')
      #
      # Because the node manager needs the group names to be uniq, but
      # for licenses it is easier to use the same id, We strip the last part
      # of the group name.
      #
      license_id = license_id && license_id.split('-')[0..-2].join('-')
      data[module_name] ||= {}
      data[module_name][license_id] = group.matching_nodes
        .sort_by {|n| n['facts_timestamp']}
        .reverse
        .collect {|n| n['certname']} if license_id
    end
  end
end
