#
# See the file "LICENSE" for the full license governing this code.
#
# frozen_string_literal: true
# encoding: UTF-8

require 'entitlement/entitlement'

module EasyType
  #
  # EasyType is a flushable provider. To use this provider, you have to
  # add certain information to the type definition.
  # You MUST define following attributes on the type
  #
  #  on_create do
  #    "create user #{self[:name]}"
  #  end
  #
  #  on_modify do
  #    "alter user #{self[:name]}"
  #  end
  #
  #  on_destroy do
  #    "drop user #{self[:name]}"
  #  end
  #
  # for all properties you MUST add
  #
  #  on_apply do
  #   "identified by #{resource[:password]}"
  #  end
  # rubocop: disable Metrics/ModuleLength
  # rubocop: disable Metrics/AbcSize
  module Provider
    attr_reader :property_flush, :property_hash

    MODULE_NAME ||= %r{^.*\/(?:modules|\w+-\w+-\w+-\w+-\w+)\/(.*)\/lib.*$}
    TYPE_NAME   ||= %r{^.*\/provider\/(.*)\/.*$}
    PROVIDER    ||= %r{^.*\/provider\/.*\/(.*)\.rb.*$}

    # @private
    def self.included(parent)
      deduct_identity
      parent.extend(ClassMethods)
      # Load the contents of the last catalog. We use this to determine of we are making a 
      # corrective or an adaptive change.
      # 
      @@history ||= Puppet::Transaction::Persistence.new
      @@history.load if @@history.data == {}
    end

    # @private
    def self.deduct_identity
      stack_entry   = caller.select { |l| l.include?('puppet/provider') }.first
      provider_dir  = File.dirname(stack_entry)
      module_name   = stack_entry.scan(MODULE_NAME).flatten.first
      type_name     = stack_entry.scan(TYPE_NAME).flatten.first
      provider_name = stack_entry.scan(PROVIDER).flatten.first
      ensure_identification(module_name, type_name, provider_name, provider_dir) if module_name && type_name
    end

    # @private
    def self.ensure_identification(module_name, type_name, provider_name, provider_dir)
      file_name = Pathname.new(provider_dir) + '.identity.rb'
      return unless identity_breached?(file_name, type_name, module_name)
      Puppet.warninging 'Module identity has been breached. Trying to repair and continue...'
      content = <<CODE
class Puppet::Type::#{type_name.capitalize}::Provider#{provider_name.capitalize}
  @type_name   = '#{type_name}'
  @module_name = '#{module_name}'
end
CODE
      replace_identity_content(file_name, content)
    end

    # @private
    def self.replace_identity_content(file_name, content)
      File.open(file_name, 'w') { |file| file.write(content) }
      FileUtils.chmod(0o755, file_name)
    end

    def self.identity_breached?(file_name, _type_name, _module_name)
      current_content = File.exist?(file_name) ? File.open(file_name, 'r', &:read) : ''
      matches = current_content.scan(/@type_name   = '(.*)'\n  @module_name = '(.*)'/).flatten
      !matches[0, 2]
    end

    # @private
    def initialize(value = {})
      super(value)
      @property_flush = {}
    end

    #
    # Checks if the resource exists. It does that by checking if the ensure property
    # in the property_hash contains :present
    #
    # @return [Boolean] true if it exsist, false if it doesn't exist
    #
    def exists?
      !@property_hash[:ensure].nil?
    end

    def to_hash
      @property_hash
    end
    #
    # Create the resource based on:
    #  - The values in the property_hash
    #  - the command set on the Type
    #  - The on_create value of the Type
    #  - The on_apply values of all the specified parameters and properties
    #
    def create
      load_identity
      ::Entitlement::Entitlement.entitled?("#{module_name}##{type_name}#create")
      if disable_corrective_ensure? && corrective_ensure?
        fail "Corrective ensure present requested by catalog, but disabled by parameter disable_corrective_ensure."
      else
        @property_flush = @resource
        @property_hash[:ensure] ||= :present
        dump(:create)
        execute_from_type('create')
        @property_flush = {}
      end
    end

    #
    # Destroy the resource based on:
    #  - the command set on the Type
    #  - The on_destroy value of the Type
    #
    def destroy
      load_identity
      ::Entitlement::Entitlement.entitled?("#{module_name}##{type_name}#destroy")
      if disable_corrective_ensure? && corrective_ensure?
        fail "Corrective ensure absent requested by catalog, but disabled by parameter disable_corrective_ensure."
      else
        dump(:destroy)
        execute_from_type('destroy')
        @property_hash.clear
        @property_flush = {}
      end
    end

    #
    # Modify the resource based on:
    #  - The values in the property_hash
    #  - the command set on the Type
    #  - The on_modify value of the Type
    #  - The on_apply values of all the specified parameters and properties
    #
    def flush
      load_identity
      ::Entitlement::Entitlement.entitled?("#{module_name}##{type_name}#update")
      return unless @property_flush && @property_flush != {}
      if disable_corrective_change? && corrective_change?
        fail "Corrective change present requested by catalog, but disabled by parameter disable_corrective_change."
      else
        dump(:modify)
        execute_from_type('modify')
      end
    end

    private

    # @private
    def dump(action)
      if ENV['EASY_TYPE_DUMP_ORIGINAL'] && action == :modify
        puts "#---- Value before #{action}  ----"
        puts resource.to_resource.to_manifest
        puts "#---- End of value before #{action} ----"
      end
      return unless ENV['EASY_TYPE_DUMP_TARGET']
      puts "#---- Value applied in #{action}  ----"
      puts resource.to_manifest
      puts "#---- End of value applied in #{action} ----"
    end

    # @private
    def execute_from_type(action)
      type_before_action(action)
      properties_before_action(action)
      type_method = resource.method("on_#{action}")
      case type_method.arity
      when 1
        with_command_builder(action, type_method)
      when 0, -1 # We allow -1 because of the RSpec expect matcher. This one changes the arity of the method
        #
        without_command_builder(action, type_method)
      else
        fail "'on_#{action}' only support with 0 or 1 argument"
      end
      properties_after_action(action)
      type_after_action(action)
    end

    #
    # Call the type action with a command_builder as parameter
    #
    def with_command_builder(action, type_method)
      command_builder = ScriptBuilder.new(:binding => resource, :acceptable_commands => resource.commands)
      line = type_method.call(command_builder)
      fail "Easy_type usage error: on_xxx methods should return a nil or a string not a #{line.class.name}" unless [String, NilClass].include?(line.class)
      command_builder.add(line)
      properties_apply_action(action, command_builder)
      command_builder.execute
    end

    #
    # Call the type action without any parameters
    #
    def without_command_builder(action, type_method)
      method, command, options = type_method.call
      #
      # TODO: Add validation of the options
      #
      append = if options.nil? || options.empty?
                 ''
               else
                 options.delete(:append)
               end
      properties = property_values(action)
      return unless command && valid_property_values(properties, options)
      properties << append
      full_command = "#{command}  #{properties.join(' ')}"
      send(method, full_command, options)
    end

    def valid_property_values(properties, options)
      !(skip_without_properties(options) && properties.empty?)
    end

    def skip_without_properties(options)
      options && options.delete(:skip_without_properties)
    end

    def properties_apply_action(action, command_builder)
      resource.properties.each do |prop|
        statement = property_apply_action(prop, action, command_builder) if should_be_in_command(prop)
        command_builder << statement unless statement == ' '
      end
      command_builder
    end

    def properties_before_action(action)
      #
      # If the action is destroy, we don't care about the value of the properties. But
      # we **do** want to ne able to activate any before_* or after_* destroy actions on the properties.
      # To facilitate this we create a new (empty) instance for all properties.
      #
      if action == 'destroy'
        resource.all_property_classes.each do |property_class|
          prop = property_class.new(:resource => resource)
          property_before_action(prop, action)
        end
      else
        resource.properties.each do |prop|
          property_before_action(prop, action) if should_be_in_command(prop)
        end
      end
    end

    def properties_after_action(action)
      #
      # If the action is destroy, we don't care about the value of the properties. But
      # we **do** want to ne able to activate any before_* or after_* destroy actions on the properties.
      # To facilitate this we create a new (empty) instance for all properties.
      #
      if action == 'destroy'
        resource.all_property_classes.each do |property_class|
          prop = property_class.new(:resource => resource)
          property_after_action(prop, action)
        end
      else
        resource.properties.each do |prop|
          property_after_action(prop, action) if should_be_in_command(prop)
        end
      end
    end

    def property_values(action)
      if action == 'destroy'
        resource.all_property_classes.each do |property_class|
          prop = property_class.new(:resource => resource)
          property_value(prop, action)
        end
        []
      else
        resource.properties.collect do |prop|
          property_value(prop, action) if should_be_in_command(prop)
        end.compact
      end
    end

    def property_value(property, action)
      method_name = "on_#{action}"
      if property.respond_to?(method_name)
        property.send(method_name)
      elsif property.respond_to?('on_apply') && %w(create modify).include?(action)
        property.on_apply
      end
    end

    def property_apply_action(property, action, command_builder)
      method_name = "on_#{action}"
      if property.respond_to?(method_name)
        property.send(method_name, command_builder)
      elsif property.respond_to?('on_apply') && %w(create modify).include?(action)
        property.on_apply(command_builder)
      end
    end

    def type_before_action(action)
      type_action('before', action)
    end

    def type_after_action(action)
      type_action('after', action)
    end

    def type_action(type, action)
      method_name = "#{type}_#{action}"
      if resource.respond_to?(method_name)
        resource.send(method_name)
      elsif resource.respond_to?("#{type}_apply")
        resource.send("#{type}_apply")
      end
    end

    #
    # Calls after_* actions for specified property
    #
    def property_after_action(property, action)
      method_name = "after_#{action}"
      if property.respond_to?(method_name)
        property.send(method_name)
      elsif property.respond_to?('after_apply') && %w(create modify).include?(action)
        property.after_apply
      end
    end

    #
    # Calls before_* actions for specified property
    #
    def property_before_action(property, action)
      method_name = "before_#{action}"
      if property.respond_to?(method_name)
        property.send(method_name)
      #
      # Only call on apply for create and modify actions
      #
      elsif property.respond_to?('before_apply') && %w(create modify).include?(action)
        property.before_apply
      end
    end

    ##
    # @private
    # Should be in command if the property has defined an apply command
    # and when it is modified e.g. in de @property_flush
    #
    def should_be_in_command(property)
      modified?(property) || in_a_modified_group?(property)
    end

    # @private
    def in_a_modified_group?(property)
      if resource.groups.include_property?(property.class)
        groups = resource.groups
        group = groups.group_for(property.class)
        properties = groups.contents_for(group)
        names = properties.map(&:name)
        is_modified = names.reduce(false) do |value, entry|
          value || @property_flush[entry]
        end
        action_defined?(property) && is_modified
      else
        false
      end
    end

    # @private
    def modified?(property)
      !@property_flush[property.name].nil?
    end

    # @private
    def action_defined?(property)
      defined?(property.on_apply) || defined?(property.on_create)
    end

    # @private
    def load_identity
      self.class.load_identity
    end


    #
    # Get the module name
    #
    def module_name
      self.class.module_name
    end

    #
    # Get the type name
    #
    def type_name
      self.class.type_name
    end

    #
    # Return true if the resource is marked for no corrective ensure. Meaning no create's
    # or absenting.
    #
    def disable_corrective_ensure?
      resource[:disable_corrective_ensure]
    end

    #
    # Return true if the resource is marked for no corrective change. Meaning no property changes
    #
    def disable_corrective_change?
      resource[:disable_corrective_change]
    end

    #
    # Corrective ensure returns true when the ensure property in manifest is correcticly changed.
    #
    def corrective_ensure?
      property_corrective_changed?(ensure_instance)
    end


    #
    # If this resource was in the catalog before, with the same vales as now, it
    # is marked as a corrective change.
    def corrective_change?
      resource.properties.any? {|p| property_corrective_changed?(p)}
    end

    #
    # Get the instance of the ensure class
    #
    def ensure_instance
      resource.properties.select {|p| p.name == :ensure}.first
    end
    #
    # This method show's if the specfied property class contains contains
    # a corrective change. This should be method on the EasyType::Property class, but because not 
    # properties include this class, we use it here.
    #
    def property_corrective_changed?(property)
      old_system_value = @@history.get_system_value(resource.ref, property.name.to_s)
      # Only idempotent properties, and cases where we have an old system_value
      # are corrective_changes.
      if property.idempotent? &&
        !property.sensitive &&
        !old_system_value.nil?

        # If the values aren't insync, we have confirmed a corrective_change
        insync = property.insync_values?(old_system_value, self.send(property.name))

        # Preserve the nil state, but flip true/false
        insync.nil? ? nil : !insync
      else
        false
      end
    end

    # nodoc
    module ClassMethods
      #
      # Get the module name
      #
      def module_name
        @module_name
      end

      #
      # Get the type name
      #
      def type_name
        @type_name
      end

      #
      # define a getter and a setter method for evert specified parameter and property in the type.
      # Define the setter so it modifies the property_hash and the property_flush based on what the
      # other provider methods expect.
      #
      def mk_resource_methods
        attributes = [resource_type.validproperties, resource_type.parameters].flatten
        raise Puppet::Error, 'no parameters or properties defined. Probably an error' if attributes == [:provider]
        attributes.each do |attr|
          attr = attr.intern
          next if attr == :name
          define_method(attr) do
            @property_hash[attr] || :absent
          end

          define_method(attr.to_s + '=') do |value|
            @property_flush[attr] = value
          end
        end
      end

      #
      # Retrieve the raw_resource information by calling the`get_raw_resources` method on the Type.
      # Map every element of this Array to a Puppet resource. and return this mapped Array
      #
      # @return [Array] Mapped Array of resources
      # @raise [Puppet::Error] When `get_raw_resources` is not defined on the type.
      #
      #
      def instances
        load_identity
        ::Entitlement::Entitlement.entitled?("#{module_name}##{type_name}#index")
        raise("information: to_get_raw_resources not defined on type #{resource_type.name}") unless defined?(resource_type.get_raw_resources)
        raw_resources = resource_type.get_raw_resources
        #
        # This is a hackish way to find out if we are in the type generation mode. If we are, we only
        # want to fetch the :ensure property. We do this to speed up auto purging on machines where
        # fetching the properties take a lot of time.
        #
        properties_to_fetch = caller.any? { |c| c.scan(/resources.rb.*generate'$/).first } ? [:ensure] : []
        raw_resources.map do |raw_resource|
          map_raw_to_resource(raw_resource, properties_to_fetch)
        end
      end
      # rubocop:enable LineLength

      #
      # Prefetch all information of the specified resource. Because we already have everything in the instances
      # array, we just have to set the provider.
      #
      # @return [Array] of Puppet Resources
      #
      def prefetch(resources)
        if resource_type.respond_to?(:execute_prefetch)
          load_identity
          ::Entitlement::Entitlement.entitled?("#{module_name}##{type_name}#index", resources)
          return resource_type.execute_prefetch(resources, self)
        end
        objects = instances
        resources.each_key do |name|
          provider = objects.find { |object| object.name == name }
          resources[name].provider = provider if provider
        end
      end
      # rubocop:enable IfUnlessModifier

      # @private
      def load_identity
        provider_name = to_s.split('::')[2].downcase
        reloaded = false
        begin
          require "puppet/provider/#{provider_name}/.identity"
        rescue LoadError => e
          raise e if reloaded
          # Use puppet plugin to download any created identity files
          `puppet plugin download`
          reloaded = true
          retry
        end
      end

      # @private
      # Creates a resource with specfied properties based on the value of property
      # values returned by the translate_to_resource methods on the parameters and
      # properies.
      #
      # The selection array allows you to fetch only a subset of the properties and parametsr
      #
      # rubocop: disable Metrics/MethodLength
      def map_raw_to_resource(raw_resource, selection = [])
        resource = {}
        selected_properties = if selection.empty?
                                Puppet.debug 'Fetching all enabled properties.'
                                non_meta_parameter_classes
                              else
                                Puppet.debug "Fetching only #{selection.join(',')} properties."
                                property_classes_for(selection) << resource_type.paramclass(:name) # Use selected classes AND name
                              end

        selected_properties.each do |parameter_class|
          next unless translation?(parameter_class)
          translate_arity = parameter_class.method(:translate_to_resource).arity
          value = case translate_arity
          when 1, -1
            parameter_class.translate_to_resource(raw_resource)
          when 2
            parameter_class.translate_to_resource(raw_resource, resource)
          else
            raise ArgumentError, "wrong number of arguments (#{translate_arity} for 1 or 2)"
          end
          #
          # Let the parameter class coerce it to the correct data type
          #
          resource[parameter_class.name] = parameter_class.coerce(value) unless value.nil?
        end
        resource[:ensure] ||= :present
        new(resource)
      end
      # rubocop: enable Metrics/MethodLength

      private

      # @private
      def translation?(parameter_class)
        defined?(parameter_class.translate_to_resource)
      end

      # @private
      def property_classes_for(properties)
        resource_type.properties.select { |c| properties.include?(c.name) }
      end

      # @private
      def non_meta_parameter_classes
        resource_type.properties + non_meta_parameters.map { |param| resource_type.paramclass(param) }
      end

      # @private
      def non_meta_parameters
        resource_type.parameters - resource_type.metaparams
      end
    end
  end
end
# rubocop: enable Metrics/ModuleLength
# rubocop: enable Metrics/AbcSize
