#
# See the file "LICENSE" for the full license governing this code.
#
# frozen_string_literal: true

# rubocop:disable Security/MarshalLoad
# rubocop:disable Style/Documentation
# rubocop:disable Metrics/ClassLength
# rubocop:disable Metrics/AbcSize
# rubocop:disable Metrics/MethodLength
# rubocop:disable Metrics/CyclomaticComplexity
# rubocop:disable Metrics/PerceivedComplexity

require 'date'
require 'facter'
require 'openssl'
require 'entitlement/nodes_entry'
require 'entitlement/custom_entry'
require 'entitlement/pay_per_use'
require 'entitlement/domains_entry'
require 'entitlement/opensource_entry'
require 'entitlement/usage_only_entry'
require 'entitlement/numbered_entry'
require 'entitlement/node_group_entry'
require 'entitlement/sized_entry'

module Entitlement
  class Entitlement
    @public_key = <<-KEY
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAw2f9dZ9REdORsXUks+ar
w2Roli6lGCbJ/eMBNPWK3lB+RKY/nPJ1PDD/o64PVPw7KgGiorehklRhFUhVrEPq
v6oscIz7xIBV9grBb042M4URZMMtrdCGHu06tMiPcz/0i/lcnrvTaLY3t8xrg8za
QzHh05GYFRn8LE8NuHcP9VZJmt2ZIUT6qZcO14MSCyqmPMKgezCrO2RCkK8Bmhce
QBiPeTcFlMdSUHOD3Lgfu7h/2VqxtRzQnk+v2qLGgmZYzdt/hml8JLwuofbCTvl2
cCeL9hqJDXahbaCjKOmkLJqpfdz22F6Kd+bpR+RQhEPPVwFdZcM1eTi/xWdOsVeo
NQIDAQAB
-----END PUBLIC KEY-----
KEY

    VALID_OPTIONS ||= [:number, :puppet_server, :modules, :from, :to, :nodes,
                     :domains, :opensource, :usage_only, :license_id, :count, :node_group, 
                     :pay_per_use, :public_key, :api_url, :filename_string, :warning_number].freeze
    FREE_MODULES  ||= ['easy_type', 'pizza'].freeze

    def self.load
      paths = Puppet.settings[:basemodulepath].split(':')
      paths << Puppet.settings[:modulepath].split(':')
      paths << "#{Puppet.settings[:environmentpath]}/#{Puppet.settings[:environment]}/modules"
      paths << Puppet.settings[:confdir]
      paths << '/etc/puppetlabs/puppet'   # Always use a default path too
      Puppet.debug "Looking for entitlement files in #{paths.join(',')}"
      entitlement_files = paths.flatten.collect { |p| Dir.glob("#{p}/**/*.entitlements") }
      if entitlement_files.flatten.empty?
        # rubocop: disable Style/EvalWithLocation
        Puppet.debug 'No entitlement files found'
        instance_eval("::Entitlement::Entitlement.define { company 'Unknown'}")
      else
        entitlement_files.flatten.each do |file|
          Puppet.debug "found entitlements file #{file}"
          next unless File.exist?(file)
          read_filename = File.basename(file)
          # rubocop: disable Style/RescueStandardError
          content = "$entitlements_file = '#{file}'\n" + decrypt(file)

          begin
            instance_eval("::Entitlement::Entitlement.define {#{content} }")
          rescue => e
            raise "Error in license. Please contact Enterprise Modules.\n #{e}"
          end
          # rubocop: enable Style/EvalWithLocation
          # rubocop: enable Style/RescueStandardError
        end
      end
    end

    def self.define(&definition)
      @instance ||= new
      @definition = definition
      read_entitlement_definition
    end

    def self.read_entitlement_definition
      @instance.instance_eval(&@definition)
      @instance.instance_eval do
        raise 'Entitlements file must contain a company name' unless @name
      end
    end

    def self.decrypt(file_name)
      content = File.read(file_name)
      if !content != ''
        Struct.new('Entitlements', :data, :iv, :key) unless defined?(Struct::Entitlements)
        entitlements = Marshal.load(content)
        public_key = OpenSSL::PKey::RSA.new(@public_key)
        cipher = OpenSSL::Cipher.new('aes-256-cbc')
        cipher.decrypt
        cipher.key = public_key.public_decrypt(entitlements.key)
        cipher.iv = public_key.public_decrypt(entitlements.iv)
        cipher.update(entitlements.data) + cipher.final
      else
        ''
      end
    end

    def self.inspect
      load unless @instance
      @instance.instance_eval do
        display_entitlements
      end
    end

    def self.purgeable
      load unless @instance
      @instance.instance_eval do
        files_to_purge = []
        files_to_keep = []
        @entitlements.each do |e|
          if e.purgable?
            files_to_purge << e.file
          else
            files_to_keep << e.file
          end
        end
        (files_to_purge.flatten - files_to_keep.flatten)
      end
    end

    def self.no_license_required?( full_action)
      module_name, resource_type, action = full_action.split('#')
      return true if virtual?(module_name)
      #
      # We want all in easy_type to be free except the validation type
      #
      return false if module_name == 'easy_type' && resource_type == 'validation'
      return true if FREE_MODULES.include?(module_name)
    end

    def self.entitled?(full_action, resources = nil)
      module_name, resource_type, action = full_action.split('#')
      return true if no_license_required?(full_action)
      load unless @instance
      @instance.instance_eval do
        Puppet.debug "Checking #{@name} entitlement for action #{action} in type #{resource_type} from module #{module_name}"
        if list_contains_valid_entitlement?(full_action, resources)
          Puppet.debug 'Entitlement ok'
          check_expiring_entitlements(full_action)
        else
          message = "#{action} in #{module_name}-#{resource_type} is not entitled on this node or domain at this point in time. Check entitlements list or purchase right entitlements."
          if action == 'index'
            Puppet.log_exception Puppet::DevError.new(message)
          else
            puts ''
            display_entitlements
            raise message
          end
        end
      end
    end

    def self.virtual?(module_name)
      if ENV['NO_VIRTUAL_ENTITLEMENT']
        false
      elsif Facter.value('is_virtual') && (Facter.value('virtual') == 'virtualbox' || (dmi = Facter.value('dmi')) && dmi.dig('product','name') == 'VirtualBox')
        @entitlement_displayed ||= {}
        unless @entitlement_displayed[module_name]
          @entitlement_displayed[module_name] = true
          Puppet.notice "Running on VM. You are free to use #{module_name} on a WorkStation VM. Check the license requirements on http://www.enterprisemodules.com/ for details."
        end
        true
      else
        false
      end
    end

    def initialize
      @entitlements          = []
    end

    def company(name)
      @name = name
    end

    def node_name
      Facter.value('fqdn').gsub(/\..*/, '')
    end

    def entitlement(entitlement, &validator)
      verify_options(entitlement)
      puppet_server   = entitlement[:puppet_server]
      node_group      = entitlement[:node_group]
      number          = entitlement[:number]
      warning_number  = entitlement[:warning_number]
      count           = entitlement[:count]
      modules         = entitlement[:modules]
      from            = entitlement[:from]
      to              = entitlement[:to]
      nodes           = entitlement[:nodes]
      domains         = entitlement[:domains]
      opensource      = entitlement[:opensource]
      license_id      = entitlement[:license_id] || 'none'
      usage_only      = entitlement[:usage_only]
      pay_per_use     = entitlement[:pay_per_use]
      public_key      = entitlement[:public_key]
      api_url         = entitlement[:api_url]
      filename_string = entitlement[:filename_string]
      entry = if count
                SizedEntry.new(modules, from, to, filename_string, count, &validator)
              elsif node_group
                NodeGroupEntry.new(modules, from, to, node_group, filename_string, number, puppet_server, license_id, warning_number, &validator)
              elsif number
                NumberedEntry.new(modules, from, to, filename_string, number, puppet_server, license_id, warning_number, &validator)
              elsif nodes
                NodesEntry.new(modules, from, to, filename_string, nodes, &validator)
              elsif domains
                DomainsEntry.new(modules, from, to, filename_string, domains, &validator)
              elsif opensource
                OpenSourceEntry.new(modules, &validator)
              elsif usage_only
                UsageOnlyEntry.new(modules, &validator)
              elsif pay_per_use
                PayPerUse.new(modules, public_key, api_url, @name, node_name, &validator)
              elsif validator
                CustomEntry.new(modules, from, to, filename_string, &validator)
              else
                raise 'invalid entitlement line in file'
              end
      @entitlements << entry
    end

    def verify_options(entitlement)
      entitlement.each_key do |o|
        raise "invalid entitlement option '#{o}'" unless VALID_OPTIONS.include?(o)
      end
    end

    def display_entitlements
      @entitlements.each do |e|
        e.inspect
        puts ''
      end
    end

    def list_contains_valid_entitlement?(module_name, resources)
      @entitlements.any? { |e| e.valid?(module_name, resources) }
    end

    def check_expiring_entitlements(module_name)
      @entitlements.each { |e| e.notify_if_expiring(module_name) }
    end
  end
end
# rubocop:enable Security/MarshalLoad
# rubocop:enable Style/Documentation
# rubocop:enable Metrics/ClassLength
# rubocop:enable Metrics/AbcSize
# rubocop:enable Metrics/MethodLength
# rubocop:enable Metrics/CyclomaticComplexity
# rubocop:enable Metrics/PerceivedComplexity
