#
# See the file "LICENSE" for the full license governing this code.
#
# frozen_string_literal: true

require 'entitlement/numbered_entry'
require 'easy_type/node_group_data'
# rubocop: disable Security/YAMLLoad
# rubocop: disable Style/RescueStandardError


module Entitlement
  #
  # This class allows you to specify the number of nodes to use the module on.
  #
  class NodeGroupEntry < NumberedEntry

    # rubocop: disable  Metrics/ParameterLists
    def initialize(modules, from, to, node_group, filename_string, number, server = nil, license_id = 'none', warn_number = nil, &proc)
      super( modules, from, to, filename_string, number, server, license_id, warn_number, &proc)
      @node_group        = node_group
    end
    # rubocop: enable  Metrics/ParameterLists

    def valid?(module_name, _resources = nil)
      begin
        @module_name = module_name.split('#').first
      rescue
        @module_name = module_name
      end
      Puppet.debug "Found a node group entitlement with id #{@license_id} allowing usage on #{@number} nodes,served from #{@registered_server}...."
      super(module_name) ? valid_node?(current_node) : false
    end

    def inspect
      super
    end

    private

    def registered_nodes
      node_data.dig(@module_name, @license_id)|| []
    end

    def registered_nodes_for(module_name, puppet_server)
      node_data.dig(module_name, @license_id) || []
    end

    #
    # Load the node group data. First we try to fetch it from the catalog. But when
    # we run this with an apply, the node group data is not in the catalog. Then we try to
    # use the cached data. But the cached node group data may not be older than half an hour.
    #
    # We create the cached data on every puppet agent run. The data from the catalog is 
    # saved to the cached node group data file
    #
    def node_data
      if in_bolt?
        ng_manager = EasyType::NodeGroupData.new('/tmp/node_groups.cache')
        data = ng_manager.load_from_cache
        data
      else
        ng_manager = EasyType::NodeGroupData.new('/var/enterprisemodules/node_groups.cache')
        if @registered_server == current_node
          ng_manager.load_from_server
        else
          begin
            data = ng_manager.load_from_catalog
            ng_manager.save_node_group_cache()
            data
          rescue Puppet::Error
            ng_manager.load_from_cache
          end
        end
      end
    end
  end
end
# rubocop: enable Security/YAMLLoad
# rubocop: enable Style/RescueStandardError
