#
# See the file "LICENSE" for the full license governing this code.
#
# frozen_string_literal: true

require 'entitlement/entry'
require 'easy_type/helpers'
require 'yaml'
require 'facter'
# rubocop: disable Security/YAMLLoad
# rubocop: disable Style/RescueStandardError


module Entitlement
  #
  # This class allows you to specify the number of nodes to use the module on.
  #
  class NumberedEntry < Entry
    include EasyType::Helpers

    def self.puppet_server
      server = if in_bolt?
        yaml_data = File.read('/tmp/run_info.yaml')
        data = Puppet::Util::Yaml.safe_load(yaml_data, [Symbol])
        data[:puppet_server]
      else
        Puppet.settings[:server]
      end
      server
    end

    ENTITLEMENTS_URL       ||= 'puppet:///em_license/entitlements.yml'.freeze
    BOLT_ENTITLEMENTS_FILE ||= '/tmp/entitlements.yml'.freeze

    # rubocop: disable  Metrics/ParameterLists
    def initialize(modules, from, to, filename_string, number, server = nil, license_id = 'none', warn_number = nil, &proc)
      super(modules, from, to, &proc)
      @number            = number
      @warn_number       = warn_number
      @licensed_server   = server || ''
      @licensed_servers  = server.split(',').map(&:strip)
      @license_id        = license_id
      @registered_server = self.class.puppet_server
    end
    # rubocop: enable  Metrics/ParameterLists

    def valid?(module_name, _resources = nil)
      begin
        @module_name = module_name.split('#').first
      rescue
        @module_name = module_name
      end
      Puppet.debug "Found a Numbered entitlement with id #{@license_id} allowing usage on #{@number} nodes,served from #{@registered_server}...."
      super(module_name) ? valid_node?(current_node) : false
    end

    def inspect
      super
      puts "License ID  : #{@license_id}"
      puts "entitled    : #{@number}"
      puts "Warn at     : #{@warn_number}" if @warn_number
      puts "on server   : #{@licensed_server}" if @licensed_server
      @modules.collect {|m| m.split('#').first}.uniq.each do | module_name|
        puts "module  : #{module_name}"
        registered_nodes = Array(registered_nodes_for(module_name, @licensed_server))
        valid            = registered_nodes.include?(current_node)
        puts "  registered             : #{registered_nodes.size}"
        puts "  on nodes               : #{registered_nodes.join(',')}"
        puts "  valid on current node  : #{valid}"
      end
    end

    def valid_node?(node)
      if puppetserver_in_hosts_file?
        Puppet.err "Entitlement doesn't support puppetserver #{@registered_server} in host file."
        return false
      end
      if invalid_puppetserver?
        Puppet.debug "License #{@license_id} is only valid on puppet server(s) #{@licensed_server}. Now running on #{@registered_server}, so skipping this one."
        return false
      end
      unless enough_nodes?
        Puppet.err "Your node list for module #{@module_name} contains more nodes then entitled. Only using first #{@number}. Please make the list compliant."
      else
        Puppet.warn_once :license, :usage_overdrawn, "Using module #{@module_name} on more then #{@warn_number} nodes. Make sure you are using the software within license limits." if above_warning_level?
      end
      unless validated_registered_nodes.include?(node)
        Puppet.debug :license, :not_registered, "Node #{node} is not registered in the node list for module #{@module_name} with license id #{@license_id}."
        return false
      else
        Puppet.debug "Node #{node} is registered in the node list for module #{@module_name} in #{ENTITLEMENTS_URL}"
      end
      Puppet.debug "Your node list for module #{@module_name} is compliant."
      true
    end

    def invalid_puppetserver?
      #
      # For now allow all puppetservers when we are using bolt. Bolt
      # cannot see which puppetserver was used.
      #
      [@registered_server] - @licensed_servers == [@registered_server]
    end

    def puppetserver_in_hosts_file?
      Resolv::Hosts.new().getaddresses(@registered_server).any?
    end

    def registered_licenses
      node_data.keys
    rescue
      []
    end

    def registered_modules
      node_data[@license_id].keys
    rescue
      []
    end

    def registered_nodes_for(module_name, puppet_server)
      node_data[@license_id][module_name][puppet_server] || []
    rescue
      []
    end

    private

    def validated_registered_nodes
      registered_nodes[0,@number]
    end

    def registered_nodes
      node_data[@license_id][@module_name][@licensed_server] || []
    rescue
      []
    end

    def node_data
      @node_data ||= YAML.load(node_list)
    end

    def enough_nodes?
      number_of_registered_nodes <= @number
    end

    def above_warning_level?
      return false if @warn_number.nil?
      number_of_registered_nodes >= @warn_number
    end

    def number_of_registered_nodes
      registered_nodes.size
    end

    def node_list
      if in_bolt?
        File.read(BOLT_ENTITLEMENTS_FILE)
      else
        begin
          get_puppet_file(ENTITLEMENTS_URL)
        rescue Net::HTTPError => e
          Puppet.warn_once(:license, :em_license_not_found, "Fileshare 'em_license' not found.\n #{e}")
          ''
        rescue Puppet::Error, ArgumentError => e
          Puppet.warn_once(:license, :no_registered_nodes, "No registered nodes found in #{ENTITLEMENTS_URL}.\n #{e}")
          ''
        end
      end
    end

    #
    # There is no safe and/or standard way to detect if we are running inside
    # of a bolt apply block. So we determine it by checking if the call stack
    # contains the term task-cache.
    #
    def self.in_bolt?
      Kernel.caller.grep(/tasks-cache/) != []
    end

    def in_bolt?
      self.class.in_bolt?
    end

    def current_node
      Facter.value('fqdn')
    end
  end
end
# rubocop: enable Security/YAMLLoad
# rubocop: enable Style/RescueStandardError
