#
# See the file "LICENSE" for the full license governing this code.
#
# frozen_string_literal: true
#
# This is a hiera data adapter that allows for version specific yaml files. Here is an example on how to use it in your hiera.yaml:
#
#       - name: "Version data for all version 12.2 versions."
#         lookup_key: 'easy_type::versioned_data'
#         path: version_12.2.yaml
#         options:
#           version_key: ibm_profile::mq_machine::version
#           specification:
#             - '>= 12.2'
#             - '< 12.3'
##
# This yaml file will loaded whith for example these versions:
#  - 12.2
#  - 12.2.1
#  - 12.2.1.1.1.1
#  - etc
#
# With the parameter 'version_key' you specify what key is used to specifiy the version you want to use to differentiate
# between multiple yaml files. 
#
# Besides a single key, you can lso use an array of version_keys. Then the first one that will resolve, will be used:
#
#       - name: "Version data for all version 12.2 versions."
#         lookup_key: 'easy_type::versioned_data'
#         path: version_12.2.yaml
#         options:
#           version_key:
#             - ora_profile::specific::version
#             - ora_profile::common::version
#           specification:
#             - '>= 12.2'
#             - '< 12.3'
#
# The specifications value can be a single string or an array of strings specificying the specification. See [here](https://guides.rubygems.org/patterns/#declaring-dependencies)
# for a full explanation
#
require 'yaml'
Puppet::Functions.create_function('easy_type::versioned_data') do

  dispatch :versioned_data do
    param 'String[1]', :lookup_key
    param 'Hash', :options
    param 'Any', :context
  end
  
  argument_mismatch :missing_path do
    param 'String[1]', :lookup_key
    param 'Hash', :options
    param 'Puppet::LookupContext', :context
  end
    
  def versioned_data(lookup_key, options, context)
    #
    # First fetch all options
    #
    path             = options['path']
    version_variable = options['version_variable']
    version_key      = options['version_key']
    specification    = options['specification']
    #
    # Now check if we may be here
    #
    return context.not_found if lookup_key == 'lookup_options' || Array(version_key).include?(lookup_key)
    #
    # Check all parameters
    #
    fail "Version specification must be specified." unless specification
    fail "Specify either 'version_key' or 'version_variable' not both." if version_key && version_variable
    fail "Specify either 'version_key' or 'version_variable'." if !version_key && !version_variable
    if version_variable
      scope = closure_scope
      unless scope.include?(version_variable)
        context.explain { "Version designating variable '#{version_variable}' not found so skipping versioned lookup."}
        return context.not_found 
      end
      value = scope[version_variable]
    else
      if version_key.is_a?(Array)
        version_key.each do | version|
          value = lookup(version)
          next if value.nil?
        end
      else
        value = lookup(version_key)
      end
      if value.nil?
        context.explain { "Version designating hiera key '#{version_key}' not found so skipping versioned lookup."}
        return context.not_found
      end
    end
    #
    # version_variable exists. No do comparisson
    #    
    unless Gem::Dependency.new('', specification).match?('', value)
      context.explain { "Version value '#{value}' not matching spec '#{specification}', so skipping versioned lookup."}
      return context.not_found 
    end
    #
    # The specified version matches to the specification. Fetch the data
    #
    data = context.cached_file_data(path) do |content|
      begin
        #
        # Try to use the safest possible way to load the yaml. This depends on 
        # the used ruby versions.
        #
        if defined?(YAML.safe_load)
          data = YAML.safe_load(content, permitted_classes: [Symbol], filename: path)
        else
          data = YAML.load(content)
        end

        if data.is_a?(Hash)
          Puppet::Pops::Lookup::HieraConfig.symkeys_to_string(data)
        else
          msg = _("%{path}: file does not contain a valid yaml hash" % { path: path })
          raise Puppet::DataBinding::LookupError, msg if Puppet[:strict] == :error && data != false
          Puppet.warning(msg)
          {}
        end
      rescue Puppet::Util::Yaml::YamlLoadError => ex
        # YamlLoadErrors include the absolute path to the file, so no need to add that
        raise Puppet::DataBinding::LookupError, _("Unable to parse %{message}") % { message: ex.message }
      end
    end
    value = data[lookup_key]
    if value.nil?
      if !data.include?(lookup_key)
        context.not_found()
      else
        return nil # nothing to process further - just return the nil value
      end
    end
    context.interpolate(value)
end

  def lookup(value)
    call_function('lookup', value, nil, 'first', nil)
  end


  def missing_path(lookup_key, options, context)
    "one of 'path', 'paths' 'glob', 'globs' or 'mapped_paths' must be declared in hiera.yaml when using this data_hash function"
  end
end